import logging

logFormatter = logging.Formatter('%(asctime)s-%(name)s:%(lineno)d - %(levelname)s - %(message)s')
rootLogger = logging.getLogger()
rootLogger.setLevel(logging.INFO)

consoleHandler = logging.StreamHandler()
consoleHandler.setLevel(logging.INFO)
consoleHandler.setFormatter(logFormatter)
rootLogger.addHandler(consoleHandler)


__version__ = "0.2.3"
__DOWNLOAD_SERVER__ = 'https://public.ukp.informatik.tu-darmstadt.de/reimers/sentence-transformers/v0.2/'
from .datasets import SentencesDataset, SentenceLabelDataset
from .data_samplers import LabelSampler
from .LoggingHandler import LoggingHandler
from .SentenceTransformer import SentenceTransformer




import gzip
import json
import logging
import os
from typing import List

import numpy as np
import torch
from torch import nn
from tqdm import tqdm

from soco_encoders.models.tokenizer import WordTokenizer, WhitespaceTokenizer
from soco_encoders.util import import_from_string, fullname, http_get


class WordEmbeddings(nn.Module):
    def __init__(self, tokenizer: WordTokenizer, embedding_weights, update_embeddings: bool = False, max_seq_length: int = 1000000):
        nn.Module.__init__(self)
        if isinstance(embedding_weights, list):
            embedding_weights = np.asarray(embedding_weights)

        if isinstance(embedding_weights, np.ndarray):
            embedding_weights = torch.from_numpy(embedding_weights)

        num_embeddings, embeddings_dimension = embedding_weights.size()
        self.embeddings_dimension = embeddings_dimension
        self.emb_layer = nn.Embedding(num_embeddings, embeddings_dimension)
        self.emb_layer.load_state_dict({'weight': embedding_weights})
        self.emb_layer.weight.requires_grad = update_embeddings
        self.tokenizer = tokenizer
        self.update_embeddings = update_embeddings
        self.max_seq_length = max_seq_length

    def forward(self, features):
        token_embeddings = self.emb_layer(features['input_ids'])
        cls_tokens = None
        features.update({'token_embeddings': token_embeddings, 'cls_token_embeddings': cls_tokens, 'input_mask': features['input_mask']})
        return features

    def get_sentence_features(self, tokens: List[str], pad_seq_length: int):
        pad_seq_length = min(pad_seq_length, self.max_seq_length)

        tokens = tokens[0:pad_seq_length] #Truncate tokens if needed
        input_ids = tokens

        sentence_length = len(input_ids)
        input_mask = [1] * len(input_ids)
        padding = [0] * (pad_seq_length - len(input_ids))
        input_ids += padding
        input_mask += padding

        assert len(input_ids) == pad_seq_length
        assert len(input_mask) == pad_seq_length

        return {'input_ids': input_ids, 'input_mask': input_mask, 'sentence_lengths': sentence_length}

        return {'input_ids': np.asarray(input_ids, dtype=np.int64),
                'input_mask': np.asarray(input_mask, dtype=np.int64),
                'sentence_lengths': np.asarray(sentence_length, dtype=np.int64)}

    def get_word_embedding_dimension(self) -> int:
        return self.embeddings_dimension

    def tokenize(self, text: str) -> List[str]:
        return self.tokenizer.tokenize(text)

    def save(self, output_path: str):
        with open(os.path.join(output_path, 'wordembedding_config.json'), 'w') as fOut:
            json.dump(self.get_config_dict(), fOut, indent=2)

        torch.save(self.state_dict(), os.path.join(output_path, 'pytorch_model.bin'))
        self.tokenizer.save(output_path)

    def get_config_dict(self):
        return {'tokenizer_class': fullname(self.tokenizer), 'update_embeddings': self.update_embeddings, 'max_seq_length': self.max_seq_length}

    @staticmethod
    def load(input_path: str):
        with open(os.path.join(input_path, 'wordembedding_config.json'), 'r') as fIn:
            config = json.load(fIn)

        tokenizer_class = import_from_string(config['tokenizer_class'])
        tokenizer = tokenizer_class.load(input_path)
        weights = torch.load(os.path.join(input_path, 'pytorch_model.bin'))
        embedding_weights = weights['emb_layer.weight']
        model = WordEmbeddings(tokenizer=tokenizer, embedding_weights=embedding_weights, update_embeddings=config['update_embeddings'])
        return model

    @staticmethod
    def from_text_file(embeddings_file_path: str, update_embeddings: bool = False, item_separator: str = " ", tokenizer=WhitespaceTokenizer(), max_vocab_size: int = None):
        logging.info("Read in embeddings file {}".format(embeddings_file_path))

        if not os.path.exists(embeddings_file_path):
            logging.info("{} does not exist, try to download from server".format(embeddings_file_path))

            if '/' in embeddings_file_path or '\\' in embeddings_file_path:
                raise ValueError("Embeddings file not found: ".format(embeddings_file_path))

            url = "https://public.ukp.informatik.tu-darmstadt.de/reimers/embeddings/"+embeddings_file_path
            http_get(url, embeddings_file_path)

        embeddings_dimension = None
        vocab = []
        embeddings = []

        with gzip.open(embeddings_file_path, "rt", encoding="utf8") if embeddings_file_path.endswith('.gz') else open(embeddings_file_path, encoding="utf8") as fIn:
            iterator = tqdm(fIn, desc="Load Word Embeddings", unit="Embeddings")
            for line in iterator:
                split = line.rstrip().split(item_separator)
                word = split[0]

                if embeddings_dimension == None:
                    embeddings_dimension = len(split) - 1
                    vocab.append("PADDING_TOKEN")
                    embeddings.append(np.zeros(embeddings_dimension))

                if (len(split) - 1) != embeddings_dimension:  # Assure that all lines in the embeddings file are of the same length
                    logging.error("ERROR: A line in the embeddings file had more or less  dimensions than expected. Skip token.")
                    continue

                vector = np.array([float(num) for num in split[1:]])
                embeddings.append(vector)
                vocab.append(word)

                if max_vocab_size is not None and max_vocab_size > 0 and len(vocab) > max_vocab_size:
                    break

            embeddings = np.asarray(embeddings)

            tokenizer.set_vocab(vocab)
            return WordEmbeddings(tokenizer=tokenizer, embedding_weights=embeddings, update_embeddings=update_embeddings)




from transformers import BertTokenizerFast
import jieba
import spacy
import json

class SocoBertTokenizer(object):

    CHAR = 'char'
    WORD = 'word'
    ALL = "all"
    UNK = '[UNK]'

    def __init__(self, base_vocab_path, final_vocab_path, lang, uncased=True, add_special_tokens=False):
        self._lang = lang
        self._uncased = uncased
        self._tokenizer = BertTokenizerFast.from_pretrained(base_vocab_path, add_special_tokens=add_special_tokens)

        if self._lang == 'en':
            self.en_spacy = spacy.load('en_core_web_sm')

        elif self._lang == 'zh':
            jieba.initialize()
            self.en_spacy = None
        else:
            raise Exception

        print("Load tokenizer {} - {}".format(base_vocab_path, lang))

        # remove words that are UNK in tokenizer's vocab
        vocab = [w if type(w) is str else tuple(w) for w in json.load(open(final_vocab_path))]
        self._valid_vocab = set(vocab)
        self.vocab2id = {v: idx for idx, v in enumerate(vocab)}
        self.id2vocab = vocab

        print("Load vocab of size (original {}) = {} ".format(self._tokenizer.vocab_size, len(vocab)))

    @property
    def sep_token(self):
        return '[SEP]'

    @property
    def cls_token(self):
        return '[CLS]'

    def _word_tokenize(self, text):
        if self._uncased:
            text = text.lower()

        if self._lang == 'en':
            return [(w.text, w.idx, w.idx+len(w)) for w in self.en_spacy(text, disable=["tagger", "parser", "ner"])]

        elif self._lang == 'zh':
            return list(jieba.tokenize(text))
        else:
            raise Exception("Not supported yet")

    def _map_word_to_subword(self, wor_tokens, sub_data):
        results = []
        for w in wor_tokens:
            if w[0] not in set(sub_data):
                results.append(w[0])

        return results

    def _normalize(self, tokens):
        new_results = []
        for t in tokens:
            if t in self._valid_vocab:
                new_results.append(t)

        return new_results

    def _tokenize(self, text, mode=CHAR):
        if mode == self.CHAR:
            results = self._tokenizer.tokenize(text)

        elif mode == self.WORD:
            wor_tokens = self._word_tokenize(text)
            results = []
            for w in wor_tokens:
                sub_tokens = self._tokenizer.tokenize(w[0])
                if len(sub_tokens) > 1 and tuple(sub_tokens) in self._valid_vocab:
                    results.append(tuple(sub_tokens))
                else:
                    results.extend(sub_tokens)

        elif mode == self.ALL:
            sub_res = self._tokenizer.tokenize(text)
            wor_tokens = self._word_tokenize(text)
            new_tokens = self._map_word_to_subword(wor_tokens, sub_res)
            results = sub_res

            for w in new_tokens:
                sub_tokens = self._tokenizer.tokenize(w)
                if len(sub_tokens) > 1 and tuple(sub_tokens) in self._valid_vocab:
                    results.append(tuple(sub_tokens))
        else:
            raise Exception("Not supported yet")

        return results

    def tokenize(self, text, mode=CHAR, normalize=True, max_len=1e10):
        if self._uncased:
            text = text.lower()

        tokens = self._tokenize(text, mode)

        max_len = int(max_len)
        if normalize:
            return self._normalize(tokens)[0:max_len]
        else:
            return tokens[0:max_len]


    def convert_tokens_to_ids(self, tokens):
        q_ids = []
        for token in tokens:
            if type(token) is str:
                q_ids.append(self.vocab2id[token])
            else:
                q_ids.append(self.vocab2id[tuple(token)])

        return q_ids

    def convert_ids_to_tokens(self, ids):
        return [self.id2vocab[id] for id in ids]







import json
import os
from typing import List

import torch
from torch import nn


class CNN(nn.Module):
    """CNN-layer with multiple kernel-sizes over the word embeddings"""

    def __init__(self, in_word_embedding_dimension: int, out_channels: int = 256, kernel_sizes: List[int] = [1, 3, 5]):
        nn.Module.__init__(self)
        self.config_keys = ['in_word_embedding_dimension', 'out_channels', 'kernel_sizes']
        self.in_word_embedding_dimension = in_word_embedding_dimension
        self.out_channels = out_channels
        self.kernel_sizes = kernel_sizes

        self.embeddings_dimension = out_channels*len(kernel_sizes)
        self.convs = nn.ModuleList()

        in_channels = in_word_embedding_dimension
        for kernel_size in kernel_sizes:
            padding_size = int((kernel_size - 1) / 2)
            conv = nn.Conv1d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size,
                             padding=padding_size)
            self.convs.append(conv)

    def forward(self, features):
        token_embeddings = features['token_embeddings']

        token_embeddings = token_embeddings.transpose(1, -1)
        vectors = [conv(token_embeddings) for conv in self.convs]
        out = torch.cat(vectors, 1).transpose(1, -1)

        features.update({'token_embeddings': out})
        return features

    def get_word_embedding_dimension(self) -> int:
        return self.embeddings_dimension

    def tokenize(self, text: str) -> List[str]:
        raise NotImplementedError()

    def save(self, output_path: str):
        with open(os.path.join(output_path, 'cnn_config.json'), 'w') as fOut:
            json.dump(self.get_config_dict(), fOut, indent=2)

        torch.save(self.state_dict(), os.path.join(output_path, 'pytorch_model.bin'))

    def get_config_dict(self):
        return {key: self.__dict__[key] for key in self.config_keys}

    @staticmethod
    def load(input_path: str):
        with open(os.path.join(input_path, 'cnn_config.json'), 'r') as fIn:
            config = json.load(fIn)

        weights = torch.load(os.path.join(input_path, 'pytorch_model.bin'))
        model = CNN(**config)
        model.load_state_dict(weights)
        return model




import torch
from torch import Tensor
from torch import nn
from typing import Union, Tuple, List, Iterable, Dict
from collections import OrderedDict
import os
import json


class MLP(nn.Module):
    """Performs MLP
    """

    def __init__(self,
                 word_embedding_dimension: int,
                 output_dimension: int,
                 num_layer: int = 2,
                 dropout: float=0.1,
                 linear_output=False
                 ):
        super(MLP, self).__init__()

        self.config_keys = ['word_embedding_dimension', 'output_dimension', 'num_layer', 'dropout', 'linear_output']

        self.word_embedding_dimension = word_embedding_dimension
        self.output_dimension = output_dimension
        self.num_layer = num_layer
        self.dropout = dropout
        self.linear_output = linear_output
        layers = OrderedDict()
        for i in range(num_layer):
            if i == 0:
                layers['fc-0'] = nn.Linear(word_embedding_dimension, output_dimension)
                layers['act-0'.format(i)] = nn.Tanh()
            else:
                layers['dropout-{}'.format(i)] = nn.Dropout(0.1)
                layers['fc-{}'.format(i)] = nn.Linear(output_dimension, output_dimension)
                layers['act-{}'.format(i)] = nn.Tanh()
        if linear_output:
            layers.pop('act-{}'.format(num_layer-1))

        print(layers)
        self.mlp = nn.Sequential(layers)

    def forward(self, features: Dict[str, Tensor]):
        sentence_embedding = features['sentence_embedding']
        new_sentence_embedding = self.mlp(sentence_embedding)
        features.update({'sentence_embedding': new_sentence_embedding})
        return features

    def get_sentence_embedding_dimension(self):
        return self.output_dimension

    def get_config_dict(self):
        return {key: self.__dict__[key] for key in self.config_keys}

    def save(self, output_path):
        with open(os.path.join(output_path, 'config.json'), 'w') as fOut:
            json.dump(self.get_config_dict(), fOut, indent=2)

        torch.save(self.state_dict(), os.path.join(output_path, 'pytorch_model.bin'))


    @staticmethod
    def load(input_path):
        with open(os.path.join(input_path, 'config.json')) as fIn:
            config = json.load(fIn)

        weights = torch.load(os.path.join(input_path, 'pytorch_model.bin'))
        model = MLP(**config)
        model.load_state_dict(weights)
        return model



from torch import Tensor
from torch import nn
from transformers import AlbertModel, AlbertTokenizer
import json
from typing import Union, Tuple, List, Dict
import os
import numpy as np
import logging

class ALBERT(nn.Module):
    """BERT model to generate token embeddings.

    Each token is mapped to an output vector from BERT.
    """
    def __init__(self, model_name_or_path: str, max_seq_length: int = 128, do_lower_case: bool = True):
        super(ALBERT, self).__init__()
        self.config_keys = ['max_seq_length', 'do_lower_case']
        self.do_lower_case = do_lower_case

        if max_seq_length > 510:
            logging.warning("BERT only allows a max_seq_length of 510 (512 with special tokens). Value will be set to 510")
            max_seq_length = 510
        self.max_seq_length = max_seq_length

        self.albert = AlbertModel.from_pretrained(model_name_or_path)
        self.tokenizer = AlbertTokenizer.from_pretrained(model_name_or_path, do_lower_case=do_lower_case)
        self.cls_token_id = self.tokenizer.convert_tokens_to_ids([self.tokenizer.cls_token])[0]
        self.sep_token_id = self.tokenizer.convert_tokens_to_ids([self.tokenizer.sep_token])[0]

    def forward(self, features):
        """Returns token_embeddings, cls_token"""
        output_tokens = self.albert(input_ids=features['input_ids'], token_type_ids=features['token_type_ids'],
                                    attention_mask=features['input_mask'])[0]
        cls_tokens = output_tokens[:, 0, :]  # CLS token is first token
        features.update({'token_embeddings': output_tokens, 'cls_token_embeddings': cls_tokens, 'input_mask': features['input_mask']})
        return features

    def get_word_embedding_dimension(self) -> int:
        return self.albert.config.hidden_size

    def tokenize(self, text: str) -> List[str]:
        """
        Tokenizes a text and maps tokens to token-ids
        """
        return self.tokenizer.convert_tokens_to_ids(self.tokenizer.tokenize(text))

    def get_sentence_features(self, tokens: List[str], pad_seq_length: int):
        """
        Convert tokenized sentence in its embedding ids, segment ids and mask

        :param tokens:
            a tokenized sentence
        :param pad_seq_length:
            the maximal length of the sequence. Cannot be greater than self.sentence_transformer_config.max_seq_length
        :return: embedding ids, segment ids and mask for the sentence
        """
        pad_seq_length = min(pad_seq_length, self.max_seq_length)

        tokens = tokens[:pad_seq_length]
        input_ids = [self.cls_token_id] + tokens + [self.sep_token_id]
        sentence_length = len(input_ids)

        pad_seq_length += 2  ##Add Space for CLS + SEP token

        token_type_ids = [0] * len(input_ids)
        input_mask = [1] * len(input_ids)

        # Zero-pad up to the sequence length. BERT: Pad to the right
        padding = [0] * (pad_seq_length - len(input_ids))
        input_ids += padding
        token_type_ids += padding
        input_mask += padding

        assert len(input_ids) == pad_seq_length
        assert len(input_mask) == pad_seq_length
        assert len(token_type_ids) == pad_seq_length

        return {'input_ids': np.asarray(input_ids, dtype=np.int64), 'token_type_ids': np.asarray(token_type_ids, dtype=np.int64), 'input_mask': np.asarray(input_mask, dtype=np.int64), 'sentence_lengths': np.asarray(sentence_length, dtype=np.int64)}

    def get_config_dict(self):
        return {key: self.__dict__[key] for key in self.config_keys}

    def save(self, output_path: str):
        self.albert.save_pretrained(output_path)
        self.tokenizer.save_pretrained(output_path)

        with open(os.path.join(output_path, 'sentence_albert_config.json'), 'w') as fOut:
            json.dump(self.get_config_dict(), fOut, indent=2)

    @staticmethod
    def load(input_path: str):
        with open(os.path.join(input_path, 'sentence_albert_config.json')) as fIn:
            config = json.load(fIn)
        return ALBERT(model_name_or_path=input_path, **config)









from soco_encoders.cloud_bucket import CloudBucket
from soco_encoders.encoders import *
from soco_encoders.Tokenizer import Tokenizer
from soco_encoders.util import Pack
from soco_device import DeviceCheck

class EncoderLoader(object):

    @staticmethod
    def _load_sent_transformer(config, model_family, model_id, use_gpu, region):
        bucket = CloudBucket(region)
        model_dir = bucket.download_transformer_model(model_family, model_id)

        constructor = globals()[config.encoder_class]
        model = constructor(model_dir, config)
        model.eval()
        if use_gpu:
            model.to(config.device)

        return model

    @staticmethod
    def _load_term_transformer(config, model_family, model_id, use_gpu, region):
        bucket = CloudBucket(region)
        model_dir = bucket.download_term_transformer_model(model_family, model_id)

        constructor = globals()[config.encoder_class]
        model = constructor(model_dir, config)
        model.eval()
        if use_gpu:
            model.to(config.device)

        return model

    @staticmethod
    def load_model(model_family, model_id, use_gpu, region='us'):
        if not torch.cuda.is_available():
            use_gpu = False

        bucket = CloudBucket(region)
        model_dir = bucket.download_config(model_family, model_id)
        config = json.load(open(os.path.join(model_dir, 'config.json')))

        device_check = DeviceCheck()
        device = device_check.get_device_by_model(model_id) if use_gpu else 'cpu'
        config['use_gpu'] = False if device == 'cpu' else True
        config['device'] = device
        config = Pack(config)
        if config.encoder_class == 'TransformerWrapper':
            return EncoderLoader._load_sent_transformer(config, model_family, model_id, use_gpu, region)
        elif config.encoder_class == 'TermTransformer':
            return EncoderLoader._load_term_transformer(config, model_family, model_id, use_gpu, region)

        else:
            raise Exception("{} is unknown.".format(config.encoder_class))

    @staticmethod
    def load_tokenizer(model_id, region='us'):
        return Tokenizer(model_id, region)




# @Time    : 9/20/17 2:15 PM
# @Author  : Tiancheng Zhao
import logging
import os
from typing import List
import json
import numpy as np
import scipy.sparse
import torch
from soco_encoders.SentenceTransformer import SentenceTransformer
from tqdm import tqdm
from soco_encoders.models.module_bases import BaseEncoder
import pickle


class TransformerWrapper(BaseEncoder):
    def __init__(self, model_dir, config):
        super(TransformerWrapper, self).__init__(None, config)
        self.model = SentenceTransformer(model_dir, device=config.device)

        print("Loaded sentence transformer at {}".format(model_dir))

    def forward(self, utts, utt_lens=None, feats=None, return_all=False):
        pass

    def get_dual_sentence_features(self, tokens, start_label, end_label, max_seq_len):
        pad_seq_length = min(max_seq_len + 2, self.model._first_module().max_seq_length)
        tokens = tokens[0:pad_seq_length - 2]

        # first example is on the START and the second example is on the END
        tokens = [self.model._first_module().cls_token_id] + tokens + [self.model._first_module().sep_token_id]
        segment_ids = [0] * len(tokens)
        input_mask = [1] * len(tokens)
        input_ids = tokens
        sentence_length = len(input_ids)

        # PAD TO MAX LEN
        padding_length = pad_seq_length - len(input_ids)
        input_ids = input_ids + ([0] * padding_length)
        input_mask = input_mask + ([0] * padding_length)
        segment_ids = segment_ids + ([0] * padding_length)

        # indicate where the answer is
        segment_ids = np.array(segment_ids)
        segment_ids[start_label:end_label] = 1
        segment_ids = segment_ids.tolist()

        assert len(input_ids) == pad_seq_length
        assert len(input_mask) == pad_seq_length
        assert len(segment_ids) == pad_seq_length

        return {'input_ids': np.asarray(input_ids, dtype=np.int64),
                'token_type_ids': np.asarray(segment_ids, dtype=np.int64),
                'input_mask': np.asarray(input_mask, dtype=np.int64),
                'sentence_lengths': np.asarray(sentence_length, dtype=np.int64)}

    def tokenize_with_brackets(self, tokenize, sent, start_symbol, end_symbol):
        prefix, tokens = sent.split(start_symbol)
        ans, postfix = tokens.split(end_symbol)

        prefix_tokens = tokenize(prefix.strip()) if prefix else []
        ans_tokens = tokenize(ans.strip())
        postfix_tokens = tokenize(postfix.strip()) if postfix else []
        tokens = prefix_tokens + ans_tokens + postfix_tokens

        s_id = len(prefix_tokens)
        ans_len = len(ans_tokens)
        # prediction labels
        start_label = s_id + 1  # CLS token
        end_label = s_id + ans_len + 1  # CLS token

        return tokens, start_label, end_label

    def advanced_encode(self, sentences: List[str], batch_size: int = 8, show_progress_bar: bool = None,
                        start_symbol: str = None, end_symbol: str = None, to_numpy: bool = True,
                        return_mask: bool = False, force_seq_len : int = -1):

        if show_progress_bar is None:
            show_progress_bar = (logging.getLogger().getEffectiveLevel() == logging.INFO
                                 or logging.getLogger().getEffectiveLevel() == logging.DEBUG)

        all_embeddings = []
        all_masks = []
        length_sorted_idx = np.argsort([len(sen) for sen in sentences])

        iterator = range(0, len(sentences), batch_size)
        if show_progress_bar:
            iterator = tqdm(iterator, desc="Batches")

        for batch_idx in iterator:

            batch_start = batch_idx
            batch_end = min(batch_start + batch_size, len(sentences))

            max_seq_len = 0
            batch_tokens = []
            for idx in length_sorted_idx[batch_start:batch_end]:
                tokens, s_id, e_id = self.tokenize_with_brackets(self.model.tokenize, sentences[idx], start_symbol,
                                                                 end_symbol)
                max_seq_len = max(max_seq_len, len(tokens))
                batch_tokens.append((tokens, s_id, e_id))

            features = {}
            for tokens, s_id, e_id in batch_tokens:
                sentence_features = self.get_dual_sentence_features(tokens, s_id, e_id, max(max_seq_len, force_seq_len))

                for feature_name in sentence_features:
                    if feature_name not in features:
                        features[feature_name] = []
                    features[feature_name].append(sentence_features[feature_name])

            for feature_name in features:
                features[feature_name] = torch.tensor(np.asarray(features[feature_name])).to(self.model.device)

            with torch.no_grad():
                embeddings = self.model.forward(features)
                embeddings = embeddings['sentence_embedding']
                mask = features['input_mask']
                if to_numpy:
                    embeddings = embeddings.to('cpu').numpy()
                    mask = mask.to('cpu').numpy()

                all_embeddings.extend(embeddings)
                all_masks.extend(mask)

            if self.use_gpu:
                with torch.cuda.device(self.model.device):
                    torch.cuda.empty_cache()

        reverting_order = np.argsort(length_sorted_idx)
        all_embeddings = [all_embeddings[idx] for idx in reverting_order]
        all_masks = [all_masks[idx] for idx in reverting_order]

        if return_mask:
            return all_embeddings, all_masks
        else:
            return all_embeddings

    def encode(self, utts, batch_size=100, show_progress_bar=False, mode=BaseEncoder.DEFAULT, args=None):
        args = dict() if args is None else args
        if mode == BaseEncoder.ANSWER:
            start_symbol = args['start_symbol']
            end_symbol = args['end_symbol']
            embeddings = self.advanced_encode(utts, batch_size=batch_size, show_progress_bar=show_progress_bar,
                                              start_symbol=start_symbol, end_symbol=end_symbol)
        elif mode == BaseEncoder.DEFAULT:
            embeddings = self.model.encode(utts, batch_size=batch_size, show_progress_bar=show_progress_bar)
        else:
            raise Exception("Unsupported mode={} for this encoder".format(mode))

        results = [np.expand_dims(feature, axis=0) for feature in embeddings]
        return np.concatenate(results, axis=0)

    def tokenize_to_ids(self, sent):
        return self.model.tokenize(sent)

    def tokenize_to_pieces(self, sent):
        token_ids = self.model.tokenize(sent)
        return self.model._first_module().tokenizer.convert_ids_to_tokens(token_ids)

    def convert_ids_to_tokens(self, token_ids):
        return self.model._first_module().tokenizer.convert_ids_to_tokens(token_ids)

    def get_embedding_dimension(self):
        return self.model.get_sentence_embedding_dimension()


class TermTransformer(TransformerWrapper):
    logger = logging.getLogger(__name__)

    def __init__(self, model_dir, config):
        super(TermTransformer, self).__init__(model_dir, config)
        self.device = config.device

        self.term_embedding = torch.load(os.path.join(model_dir, 'term_embeddings.bin'), map_location=self.device)
        if os.path.isfile(os.path.join(model_dir, 'term_vocab.pkl')):
            self.term_vocab = pickle.load(open(os.path.join(model_dir, 'term_vocab.pkl'),"rb"))
        else:
            self.term_vocab = json.load(open(os.path.join(model_dir, 'term_vocab.json')))

        if config.use_gpu:
            self.logger.info("Use pytorch device: {}".format(self.device))
            self.term_embedding.to(self.device)

        self.log_scale = config.log_scale
        self.score_bias = config.get('score_bias', 0)
        self.max_attention = config.get('max_attention', False)
        self.tokenizer_id = config.get('tokenizer_id', None)
        self.query_tokenize_mode = config.get('query_tokenize_mode', 'char')

    def _multi_head_cross_similarity(self, q_feature, c_feature, q_mask, c_mask, mask_padding=False, approx_padding=False):
        """
        :param q_feature: query_size x T x hidden_size
        :param c_feature: doc_size x num_head x hidden_size
        :param q_mask: query_size x T
        :param c_mask: doc_size x num_token
        :return:
        """
        Q = q_feature.shape[0]
        T = q_feature.shape[1]
        H = q_feature.shape[2]
        A = c_feature.shape[0]
        N = c_feature.shape[1]

        flat_c_feature = c_feature.view(-1, H)
        flat_q_feature = q_feature.view(-1, H)

        head_logits = torch.matmul(flat_q_feature, flat_c_feature.transpose(0, 1))
        logits = head_logits.view(-1, A, N)  # (Q x T) x A x N

        if mask_padding:
            flat_q_mask = q_mask.view(-1, 1)
            flat_c_mask = c_mask[:, 0:N].view(-1, 1)
            logit_mask = torch.matmul(flat_q_mask, flat_c_mask.transpose(0, 1))
            logit_mask = (1.0 - logit_mask) * -10000.0
            logit_mask = logit_mask.view(-1, A, N)
            logits += logit_mask

        if self.max_attention:
            max_logitis, _ = torch.max(logits, dim=2)
            match_logitis = max_logitis.view(Q, T, A)
        else:
            probs = torch.softmax(logits, dim=2).unsqueeze(2)  # (QxT)x A x 1 x N
            new_c_feature = torch.matmul(probs, c_feature).squeeze(2)  # (QxT) x A x H

            new_c_feature = new_c_feature.view(Q, T, A, H)  # batch_size x T x doc_size x hidden_size
            match_logitis = torch.matmul(q_feature.unsqueeze(2), new_c_feature.transpose(2, 3)).squeeze(2)

            if approx_padding:
                seq_len = torch.sum(c_mask, dim=1)
                match_logitis += torch.clamp((seq_len * (1 / 320) - 1).view(1, 1, -1), max=0.0)

        match_logitis = match_logitis + self.score_bias

        match_logitis = torch.relu(match_logitis)
        if self.log_scale:
            match_logitis = torch.log(match_logitis + 1.0)

        match_logitis = torch.sum(match_logitis, dim=1)
        return match_logitis

    def _top_p(self, matrix, p):
        """
        :param matrix: num_sent x num_terms
        :param p: [0 to 1]
        :return:
        """
        sorted_matrix = np.sort(matrix * -1, axis=1) * -1
        p_matrix = np.cumsum(sorted_matrix, axis=1) / np.sum(matrix, axis=1, keepdims=True)
        result = np.array(matrix, copy=True)
        for row_id in range(matrix.shape[0]):
            th = 0
            for col_id in range(matrix.shape[1]):
                th = sorted_matrix[row_id, col_id]
                if p_matrix[row_id, col_id] >= p:
                    break

            row = matrix[row_id]
            row[row < th] = 0.0
            result[row_id] = row

        return result

    def _top_k(self, matrix, k):
        sorted_matrix = np.sort(matrix * -1, axis=1) * -1
        result = np.array(matrix, copy=True)
        for row_id in range(matrix.shape[0]):
            th = sorted_matrix[row_id][k]

            row = matrix[row_id]
            row[row < th] = 0.0
            result[row_id] = row

        return result

    def _clamp(self, matrix, min_threshold):
        """
        :param matrix: num_sent x num_terms
        :param p: [0 to 1]
        :return:
        """
        matrix[matrix <= min_threshold] = 0.0
        return matrix

    def encode(self, sentences, batch_size=100, show_progress_bar=False, mode=BaseEncoder.TSCORE, args=None):

        if mode != BaseEncoder.TSCORE:
            raise Exception("Unsupported mode={} for this encoder".format(mode))

        args = dict() if args is None else args
        start_symbol = args.get('start_symbol', None)
        end_symbol = args.get('end_symbol', None)
        top_p = args.get('top_p', None)
        top_k = args.get('top_k', None)
        min_threshold = args.get('min_threshold', None)
        return_meta = args.get('return_meta', False)
        to_sparse = args.get('to_sparse', None)
        term_batch_size = args.get('term_batch_size', 1000)
        force_seq_len = args.get('force_seq_len', -1)
        mask_padding = args.get('mask_padding', True)
        approx_padding = args.get('approx_padding', False)

        vocab_size = self.term_embedding.num_embeddings
        all_term_scores = []
        length_sorted_idx = np.argsort([len(sen) for sen in sentences])

        iterator = range(0, len(sentences), batch_size)
        if show_progress_bar:
            iterator = tqdm(iterator, desc="Batches")

        with torch.no_grad():

            for batch_idx in iterator:
                batch_start = batch_idx
                batch_end = min(batch_start + batch_size, len(sentences))
                batch = [sentences[idx] for idx in length_sorted_idx[batch_start:batch_end]]

                if start_symbol is not None and end_symbol is not None:
                    embeddings, masks = self.advanced_encode(batch, batch_size=batch_size, show_progress_bar=False,
                                                             start_symbol=start_symbol, end_symbol=end_symbol,
                                                             to_numpy=False, return_mask=True,
                                                             force_seq_len=force_seq_len)

                    embeddings = torch.cat([e.unsqueeze(0) for e in embeddings], dim=0)
                    masks = torch.cat([e.unsqueeze(0) for e in masks], dim=0)
                else:
                    raise Exception("Currently default encoding is not supported yet")

                # compute term wise weights
                s_id = 0
                e_id = min(s_id + term_batch_size, vocab_size)
                temp_loigits = []
                while True:
                    terms = torch.arange(s_id, e_id).long().to(self.device)
                    term_vectors = self.term_embedding(terms).unsqueeze(1)
                    temp = self._multi_head_cross_similarity(term_vectors, embeddings,
                                                             q_mask=torch.ones(term_vectors.size()[0:2]).to(self.device),
                                                             c_mask=masks.float(),
                                                             mask_padding=mask_padding,
                                                             approx_padding = approx_padding
                                                             )
                    temp_loigits.append(temp.detach().cpu().transpose(0, 1).numpy())
                    if e_id >= vocab_size:
                        break

                    s_id = e_id
                    e_id = min(s_id + term_batch_size, vocab_size)

                temp_loigits = np.concatenate(temp_loigits, axis=1)
                all_term_scores.append(temp_loigits)

        reverting_order = np.argsort(length_sorted_idx)
        all_term_scores = np.concatenate(all_term_scores, axis=0)
        all_term_scores = [all_term_scores[idx] for idx in reverting_order]
        all_term_scores = np.array(all_term_scores)
        nz_count = np.count_nonzero(all_term_scores)
        logging.info("Encoding non-zero rate {:.3f}".format(nz_count / float(all_term_scores.size)))

        if top_p is not None:
            all_term_scores = self._top_p(all_term_scores, top_p)
            nz_count = np.count_nonzero(all_term_scores)
            self.logger.info("Top p={} non-zero rate {:.3f}".format(top_p, nz_count / float(all_term_scores.size)))

        if top_k is not None:
            all_term_scores = self._top_k(all_term_scores, top_k)
            nz_count = np.count_nonzero(all_term_scores)
            self.logger.info("Top k={} non-zero rate {:.3f}".format(top_k, nz_count / float(all_term_scores.size)))

        if min_threshold is not None:
            all_term_scores = self._clamp(all_term_scores, min_threshold)
            nz_count = np.count_nonzero(all_term_scores)
            self.logger.info(
                "Min threshold={} non-zero rate {:.3f}".format(min_threshold, nz_count / float(all_term_scores.size)))

        if to_sparse is not None:
            if to_sparse == 'csr':
                all_term_scores = scipy.sparse.csr_matrix(all_term_scores)
            elif to_sparse == 'coo':
                all_term_scores = scipy.sparse.coo_matrix(all_term_scores)
            elif to_sparse == 'dok':
                all_term_scores = scipy.sparse.dok_matrix(all_term_scores)
            else:
                self.logger.warning("Unknown sparse format {}".format(to_sparse))

        if return_meta:
            return all_term_scores, {'vocab': self.term_vocab}
        else:
            return all_term_scores




from soco_encoders.models.tokenizer.SocoBertTokenizer import SocoBertTokenizer
from soco_encoders.cloud_bucket import CloudBucket

import os
import json


class Tokenizer(object):
    local_dir = "resources"

    def __init__(self, model_id, region):
        # check if model path is on disk. o/w download it
        bucket = CloudBucket(region)
        bucket.download_tokenizer('tokenizers', model_id, local_dir=self.local_dir)
        if os.path.exists(os.path.join(self.local_dir, model_id, 'vocab.txt')):
            base_path = os.path.join(self.local_dir, model_id, 'vocab.txt'.format(model_id))
        else:
            base_path = os.path.join(self.local_dir, model_id, '{}.txt'.format(model_id))

        final_path = os.path.join(self.local_dir, model_id, '{}.json'.format(model_id))
        config_path = os.path.join(self.local_dir, model_id, 'config.json')
        config = json.load(open(config_path, 'r'))

        # create a tokenizer
        self._tokenizer = SocoBertTokenizer(base_path, final_path, config['lang'])

    def tokenize(self, *args, **kwargs):
        return self._tokenizer.tokenize(*args, **kwargs)

    def convert_tokens_to_ids(self, *args):
        return self._tokenizer.convert_tokens_to_ids(*args)

    def convert_ids_to_tokens(self, *args):
        return self._tokenizer.convert_ids_to_tokens(*args)

if __name__ == "__main__":
    t = Tokenizer('bert-base-chinese', 'cn')
    sent = '我爱自然语言处理。什么！谁敢挑战我？'
    print(t.tokenize(sent, mode='char'))
    print(t.tokenize(sent, mode='word'))
    print(t.tokenize(sent, mode='all'))
    print(t.convert_tokens_to_ids(t.tokenize(sent, mode='all')))
    print(t.convert_ids_to_tokens(t.convert_tokens_to_ids(t.tokenize(sent, mode='all'))))



import torch
from temp_model import TermQAModel
import json
import os
from soco_encoders.utils import Pack
import time


model_dir = 'resources/bert-base-uncase-ti-log-320head-s_n'
config = json.load(open(os.path.join(model_dir, 'config.json')))
config['use_gpu'] = torch.cuda.is_available()
config = Pack(config)
model = TermQAModel(model_dir, config)
model.eval()
q_batch_size = 1000
a_batch_size = 10
q_ids = torch.arange(1000, 1000+q_batch_size).long().view(q_batch_size, 1)
a_input_ids = torch.ones(a_batch_size, 320).long() * 1233
a_attention_mask = torch.ones(a_batch_size, 320).long()
a_token_type_ids = torch.ones(a_batch_size, 320).long()


inputs = {
    'q_ids': q_ids,
    'a_input_ids': a_input_ids,
    'a_attention_mask': a_attention_mask,
    'a_token_type_ids': a_token_type_ids
}

output_model_path = './bert-base-uncase-ti-log-320head-s_n.onnx'
outputs = {}

with torch.no_grad():
    inputs = {k: v.to('cuda') if config.use_gpu else v for k, v in inputs.items()}
    start = time.time()
    model(**inputs)
    end = time.time()
    print("PyTorch Inference time = ", end - start)

import onnxruntime as rt

sess_options = rt.SessionOptions()

# Set graph optimization level to ORT_ENABLE_EXTENDED to enable bert optimization.
sess_options.graph_optimization_level = rt.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
sess_options.intra_op_num_threads=8

# To enable model serialization and store the optimized graph to desired location.
sess_options.optimized_model_filepath = os.path.join('pytorch_output',
                                                     "optimized_bert-base-uncase-ti-log-320head-s_n.onnx")
session = rt.InferenceSession(output_model_path, sess_options)

# evaluate the model
rt_input =  {
    'q_ids': inputs['q_ids'].cpu().numpy(),
    'a_input_ids': inputs['a_input_ids'].cpu().numpy(),
    'a_attention_mask': inputs['a_attention_mask'].cpu().numpy(),
    'a_token_type_ids': inputs['a_token_type_ids'].cpu().numpy()
}
start = time.time()
res = session.run(None, rt_input)
end = time.time()
print("ONNX Runtime inference time: ", end - start)


from torch import Tensor
import json
import logging
import os
from typing import List, Dict

import numpy as np
from torch import Tensor
from torch import nn

from soco_encoders.models.tokenizer import WhitespaceTokenizer


class BoW(nn.Module):
    """Implements a Bag-of-Words (BoW) model to derive sentence embeddings.

    A weighting can be added to allow the generation of tf-idf vectors. The output vector has the size of the vocab.
    """

    def __init__(self, vocab: List[str], word_weights: Dict[str, float] = {}, unknown_word_weight: float = 1, cumulative_term_frequency: bool = True):
        super(BoW, self).__init__()
        vocab = list(set(vocab)) #Ensure vocab is unique
        self.config_keys = ['vocab', 'word_weights', 'unknown_word_weight', 'cumulative_term_frequency']
        self.vocab = vocab
        self.word_weights = word_weights
        self.unknown_word_weight = unknown_word_weight
        self.cumulative_term_frequency = cumulative_term_frequency

        #Maps wordIdx -> word weight
        self.weights = []
        num_unknown_words = 0
        for word in vocab:
            weight = unknown_word_weight
            if word in word_weights:
                weight = word_weights[word]
            elif word.lower() in word_weights:
                weight = word_weights[word.lower()]
            else:
                num_unknown_words += 1
            self.weights.append(weight)

        logging.info("{} out of {} words without a weighting value. Set weight to {}".format(num_unknown_words, len(vocab), unknown_word_weight))

        self.tokenizer = WhitespaceTokenizer(vocab, stop_words=set(), do_lower_case=False)
        self.sentence_embedding_dimension = len(vocab)


    def forward(self, features: Dict[str, Tensor]):
        #Nothing to do, everything is done in get_sentence_features
        return features

    def tokenize(self, text: str) -> List[str]:
        return self.tokenizer.tokenize(text)

    def get_sentence_embedding_dimension(self):
        return self.sentence_embedding_dimension

    def get_sentence_features(self, tokens: List[str], pad_seq_length: int):
        #return {'input_ids': tokens}
        vector = np.zeros(self.get_sentence_embedding_dimension(), dtype=np.float32)
        for token in tokens:
            if self.cumulative_term_frequency:
                vector[token] += self.weights[token]
            else:
                vector[token] = self.weights[token]

        return {'sentence_embedding': vector}

    def get_config_dict(self):
        return {key: self.__dict__[key] for key in self.config_keys}

    def save(self, output_path):
        with open(os.path.join(output_path, 'config.json'), 'w') as fOut:
            json.dump(self.get_config_dict(), fOut, indent=2)

    @staticmethod
    def load(input_path):
        with open(os.path.join(input_path, 'config.json')) as fIn:
            config = json.load(fIn)

        return BoW(**config)


# @Time    : 9/20/17 2:10 PM
# @Author  : Tiancheng Zhao
import torch.nn as nn


class BaseEncoder(nn.Module):
    DEFAULT = "default"
    ANSWER = "answer"
    TSCORE = "tscore"

    def __init__(self, vocab, config):
        super(BaseEncoder, self).__init__()
        self.vocab = vocab
        self.config = config
        self.use_gpu = config.use_gpu

    def encode(self, *args, **kwargs):
        raise NotImplementedError

    def tokenize_to_ids(self, sent):
        raise NotImplementedError

    def tokenize_to_pieces(self, sent):
        raise NotImplementedError

    def advanced_encode(self, *args, **kwargs):
        pass

    def get_embedding_dimension(self):
        raise NotImplementedError



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="soco-encoders",
    version="0.2.7",
    author="tianchez",
    description="Neural encoders for embedding diverse objects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.soco.ai",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Free for non-commercial use",
        "Operating System :: OS Independent",
    ],
    install_requires = [
        'torch >= 1.1.0',
        'numpy >= 1.15.3',
        'boto3 >= 1.9.46',
        'oss2 >= 2.6.0',
        'scipy >= 1.2.1',
        'transformers == 2.8.0',
        'tqdm >= 4.45.0',
        'scikit-learn >= 0.22.2.post1',
        'nltk >= 3.5',
        'jieba >= 0.42.1',
        'spacy >= 2.2.4',
        'requests >= 2.23.0'
    ]
)


import requests
from torch import Tensor, device
import torch
from tqdm import tqdm
import sys
import importlib
import numpy as np
import re

def batch_to_device(batch, target_device: device):
    """
    send a batch to a device

    :param batch:
    :param target_device:
    :return: the batch sent to the device
    """
    features = batch['features']
    for paired_sentence_idx in range(len(features)):
        for feature_name in features[paired_sentence_idx]:
            features[paired_sentence_idx][feature_name] = features[paired_sentence_idx][feature_name].to(target_device)

    labels = batch['labels'].to(target_device)
    return features, labels


def http_get(url, path):
    file_binary = open(path, "wb")
    req = requests.get(url, stream=True)
    if req.status_code != 200:
        print("Exception when trying to download {}. Response {}".format(url, req.status_code), file=sys.stderr)
        req.raise_for_status()

    content_length = req.headers.get('Content-Length')
    total = int(content_length) if content_length is not None else None
    progress = tqdm(unit="B", total=total, unit_scale=True)
    for chunk in req.iter_content(chunk_size=1024):
        if chunk: # filter out keep-alive new chunks
            progress.update(len(chunk))
            file_binary.write(chunk)
    progress.close()


def fullname(o):
  # o.__module__ + "." + o.__class__.__qualname__ is an example in
  # this context of H.L. Mencken's "neat, plausible, and wrong."
  # Python makes no guarantees as to whether the __module__ special
  # attribute is defined, so we take a more circumspect approach.
  # Alas, the module name is explicitly excluded from __qualname__
  # in Python 3.

  module = o.__class__.__module__
  if module is None or module == str.__class__.__module__:
    return o.__class__.__name__  # Avoid reporting __builtin__
  else:
    return module + '.' + o.__class__.__name__


def import_from_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        msg = "%s doesn't look like a module path" % dotted_path
        raise ImportError(msg)

    module = importlib.import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        msg = 'Module "%s" does not define a "%s" attribute/class' % (module_path, class_name)
        raise ImportError(msg)


class Pack(dict):
    def __getattr__(self, name):
        return self[name]

    def clone_dict(self, x):
        for k, v in list(x.items()):
            self[k] = v

    def add(self, **kwargs):
        for k, v in list(kwargs.items()):
            self[k] = v

    def copy(self):
        pack = Pack()
        for k, v in list(self.items()):
            if type(v) is list:
                pack[k] = list(v)
            else:
                pack[k] = v
        return pack


class ZhNumberConverter(object):
    common_used_numerals_tmp = {'零': 0, '一': 1, '二': 2, '两': 2, '三': 3, '四': 4, '五': 5, '六': 6, '七': 7, '八': 8, '九': 9,
                                '十': 10, '百': 100, '千': 1000, '万': 10000, '亿': 100000000}
    common_used_numerals = {}
    for key in common_used_numerals_tmp:
        common_used_numerals[key] = common_used_numerals_tmp[key]

    def chinese2digits(self, uchars_chinese):
        total = 0
        r = 1  # 表示单位：个十百千...
        for i in range(len(uchars_chinese) - 1, -1, -1):
            val = self.common_used_numerals.get(uchars_chinese[i])
            if val >= 10 and i == 0:  # 应对 十三 十四 十*之类
                if val > r:
                    r = val
                    total = total + val
                else:
                    r = r * val
                    # total =total + r * x
            elif val >= 10:
                if val > r:
                    r = val
                else:
                    r = r * val
            else:
                total = total + r * val
        return total

    num_str_start_symbol = ['一', '二', '两', '三', '四', '五', '六', '七', '八', '九', '十']
    more_num_str_symbol = ['零', '一', '二', '两', '三', '四', '五', '六', '七', '八', '九', '十', '百', '千', '万', '亿']

    def changeChineseNumToArab(self, oriStr):
        lenStr = len(oriStr)
        aProStr = ''
        if lenStr == 0:
            return aProStr

        hasNumStart = False
        numberStr = ''
        for idx in range(lenStr):
            if oriStr[idx] in self.num_str_start_symbol:
                if not hasNumStart:
                    hasNumStart = True

                numberStr += oriStr[idx]
            else:
                if hasNumStart:
                    if oriStr[idx] in self.more_num_str_symbol:
                        numberStr += oriStr[idx]
                        continue
                    else:
                        numResult = str(self.chinese2digits(numberStr))
                        numberStr = ''
                        hasNumStart = False
                        aProStr += numResult

                aProStr += oriStr[idx]
                pass

        if len(numberStr) > 0:
            resultNum = self.chinese2digits(numberStr)
            aProStr += str(resultNum)

        return aProStr


def set_seed(seed):
    """Sets random seed everywhere."""
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
    np.random.seed(seed)


def norm_sent(text):
    if text is None:
        return ''

    text = re.sub('\s+', ' ', text).strip()
    text = text.replace('\n', '').lower()
    text = re.sub('<[^<]+?>', '', text)
    return text



from .BERT import BERT
from .BoW import BoW
from .CNN import CNN
from .Dense import Dense
from .LSTM import LSTM
from .Pooling import Pooling
from .RoBERTa import RoBERTa
from .WordEmbeddings import WordEmbeddings
from .WordWeights import WordWeights
from .XLNet import XLNet
from .ALBERT import ALBERT
from .MLP import MLP



import boto3
import oss2
import os
import time


class CloudBucket(object):
    roles = {'reader-s3': ('AKIAIXQ4BV6ORFZ3JKPA', 'D+9W2Zii2Hvpo6G0IJ77tofKUM59ZasC/HE3Kf/w'),
             'full-s3': ('AKIAIRZZT5BICC5NFNOA', 'HB2ixc2F12xaFEuqN1G7HwP4VvfgB+l/kgxNXLFd'),
             'reader-oss': ('LTAINYtHd5knNlAg', 'HIg89gNwNYpbEKoC38SdnYCn3Db8cI'),
             'full-oss': ('LTAIkryioiCP2Wn9', 'cutsfNk7rY05qhBVNikYkxny8I4UkK')}

    def __init__(self, region='us', permission='reader'):
        self.region = region.lower()
        self.permission = permission.lower()
        if permission == 'reader':
            s3_key = self.roles['reader-s3']
            oss_key = self.roles['reader-oss']
        elif permission == 'full':
            s3_key = self.roles['full-s3']
            oss_key = self.roles['full-oss']
        else:
            raise Exception("Unknown permission {}".format(permission))
        self._s3 = boto3.resource('s3', aws_access_key_id=s3_key[0], aws_secret_access_key=s3_key[1])
        self._s3_bucket = self._s3.Bucket('convmind-models')
        self._oss_bucket = oss2.Bucket(oss2.Auth(oss_key[0], oss_key[1]),
                                       str('http://oss-accelerate.aliyuncs.com'), str('convmind-models'))

    def safe_mkdir(self, path):
        if not os.path.exists(path):
            os.mkdir(path)

    def download(self, folder_dir, files, local_dir):
        start = time.time()
        for f in files:
            src = '{}/{}'.format(folder_dir, f)
            dest =  os.path.join(local_dir, f)
            try:
                if self.region.lower() == 'cn':
                    self._oss_bucket.get_object_to_file(src, dest)
                else:
                    self._s3_bucket.download_file(src, dest)
            except Exception as e:
                print(e)
                print("Failed to download {}".format(src))

        end = time.time()
        print(("download models using {:.4f} sec".format(end - start)))

    def download_dir(self, folder_dir, dirs, local_dir):
        start = time.time()
        for d in dirs:
            cur_remote_dir = os.path.join(folder_dir, d)
            cur_local_dir = os.path.join(local_dir, d)

            if not os.path.exists(cur_local_dir):
                os.mkdir(cur_local_dir)

            files = []
            if self.region.lower() == 'cn':
                for o in self._oss_bucket.list_objects(prefix=cur_remote_dir).object_list:
                    f_name = o.key.split('/')[-1]
                    if f_name != '':
                        files.append(o.key.split('/')[-1])
            else:
                for o in self._s3_bucket.objects.filter(Prefix=cur_remote_dir):
                    f_name = o.key.split('/')[-1]
                    if f_name != '':
                        files.append(o.key.split('/')[-1])

            self.download(cur_remote_dir, files, cur_local_dir)

        end = time.time()
        print(("download models using {:.4f} sec".format(end - start)))

    def download_config(self, asset_dir, asset_id, local_dir='resources'):
        self.safe_mkdir(local_dir)

        files = ['config.json']
        model_dir = '{}/{}'.format(local_dir, asset_id)
        if not os.path.exists(model_dir):
            self.safe_mkdir(model_dir)

            self.download('{}/{}'.format(asset_dir, asset_id), files, 'resources/{}'.format(asset_id))
        else:
            print("Download is canceled since the asset already exists.")

        return model_dir

    def download_model(self, asset_dir, asset_id, local_dir='resources'):
        self.safe_mkdir(local_dir)

        files = ['m.model', 'm.vocab', 'model', 'params.json']
        model_dir = '{}/{}'.format(local_dir, asset_id)
        if not os.path.exists(os.path.join(model_dir, files[0])):
            self.safe_mkdir(model_dir)
            self.download('{}/{}'.format(asset_dir, asset_id), files, 'resources/{}'.format(asset_id))
        else:
            print("Download is canceled since the asset already exists.")

        return model_dir

    def download_transformer_model(self, asset_dir, asset_id, local_dir='resources'):
        self.safe_mkdir(local_dir)

        files = ['modules.json']
        dirs = ['0_BERT', '1_Pooling']

        model_dir = '{}/{}'.format(local_dir, asset_id)
        if not os.path.exists(os.path.join(model_dir, files[0])):
            self.safe_mkdir(model_dir)

            self.download('{}/{}'.format(asset_dir, asset_id), files, 'resources/{}'.format(asset_id))
            self.download_dir('{}/{}'.format(asset_dir, asset_id), dirs, 'resources/{}'.format(asset_id))
        else:
            print("Download is canceled since the asset already exists.")

        return model_dir

    def download_term_transformer_model(self, asset_dir, asset_id, local_dir='resources'):
        self.safe_mkdir(local_dir)

        files = ['modules.json', 'term_embeddings.bin', 'term_vocab.json']
        dirs = ['0_BERT', '1_Pooling']

        model_dir = '{}/{}'.format(local_dir, asset_id)
        if not os.path.exists(os.path.join(model_dir, files[0])):
            self.safe_mkdir(model_dir)

            self.download('{}/{}'.format(asset_dir, asset_id), files, 'resources/{}'.format(asset_id))
            self.download_dir('{}/{}'.format(asset_dir, asset_id), dirs, 'resources/{}'.format(asset_id))
        else:
            print("Download is canceled since the asset already exists.")

        return model_dir

    def download_tokenizer(self, asset_dir, asset_id, local_dir="resources"):
        self.safe_mkdir(local_dir)
        files = ['vocab.txt', '{}.txt'.format(asset_id), '{}.json'.format(asset_id), 'config.json']
        model_dir = '{}/{}'.format(local_dir, asset_id)
        if not os.path.exists(os.path.join(model_dir, files[0])):
            self.safe_mkdir(model_dir)
            self.download('{}/{}'.format(asset_dir, asset_id), files, 'resources/{}'.format(asset_id))
        else:
            print("Download is canceled since the asset already exists.")

        return model_dir



if __name__ == '__main__':
    b = CloudBucket('us')
    b.download_tokenizer('tokenizers', 'bert-base-chinese')


import torch
import os
import logging
from soco_sentence_transformers import SentenceTransformer


class TermQAModel(torch.nn.Module):
    logger = logging.getLogger(__name__)

    def __init__(self, model_dir, config):
        super(TermQAModel, self).__init__()
        self.term_embedding = torch.load(os.path.join(model_dir, 'term_embeddings.bin'))
        self.a_bert = SentenceTransformer(model_dir, device='cuda' if config.use_gpu else 'cpu')
        self.device = "cuda" if torch.cuda.is_available() and config.use_gpu else "cpu"

        if config.use_gpu:
            self.logger.info("Use pytorch device: {}".format(self.device))
            self.term_embedding.to(self.device)

    def _multi_head_cross_similarity(self, q_feature, c_feature):
        """
        :param q_feature: query_size x T x hidden_size
        :param c_feature: doc_size x num_head x hidden_size
        :param q_mask: query_size x T
        :return:
        """
        Q = q_feature.shape[0]
        T = q_feature.shape[1]
        H = q_feature.shape[2]
        A = c_feature.shape[0]
        N = c_feature.shape[1]

        flat_c_feature = c_feature.view(-1, H) # (A x N) x H
        flat_q_feature = q_feature.view(-1, H) # (Q x T) x H

        head_logits = torch.matmul(flat_q_feature, flat_c_feature.transpose(0, 1))
        logits = head_logits.view(-1, A, N)  # (QxT) x A x N

        probs = torch.softmax(logits, dim=2).unsqueeze(2)  # (QxT)x A x 1 x N
        # (QxT) x A x H
        new_c_feature = torch.matmul(probs, c_feature).squeeze(2)

        new_c_feature = new_c_feature.view(Q, T, A, H)  # batch_size x T x doc_size x hidden_size

        match_logitis = torch.matmul(q_feature.unsqueeze(2), new_c_feature.transpose(2, 3)).squeeze(2)

        match_logitis = torch.log(torch.relu(match_logitis) + 1.0)
        match_logitis = torch.sum(match_logitis, dim=1)
        return match_logitis

    def encode_ans(self, a_input_ids, a_token_type_ids, a_attention_mask):
        a_outs = self.a_bert({'input_ids': a_input_ids, 'token_type_ids': a_token_type_ids, 'input_mask': a_attention_mask})
        a_embedded = a_outs['sentence_embedding']
        return a_embedded

    def encode_que(self, q_ids):
        return self.term_embedding(q_ids)

    def forward(self, q_ids, a_input_ids, a_attention_mask, a_token_type_ids):
        q_embedded = self.encode_que(q_ids)
        a_embedded = self.encode_ans(a_input_ids, a_token_type_ids, a_attention_mask)

        return self._multi_head_cross_similarity(q_embedded, a_embedded)



import torch
from temp_model import TermQAModel
import json
import os
from soco_encoders.utils import Pack
import time
import onnxmltools
from onnxmltools.utils.float16_converter import convert_float_to_float16

model_dir = 'resources/bert-base-uncase-ti-log-320head-s_n'
config = json.load(open(os.path.join(model_dir, 'config.json')))
config['use_gpu'] = torch.cuda.is_available()
config = Pack(config)
model = TermQAModel(model_dir, config)
model.eval()
q_batch_size = 1000
a_batch_size = 1
q_ids = torch.arange(1000, 1000+q_batch_size).long().view(q_batch_size, 1)
a_input_ids = torch.ones(a_batch_size, 320).long() * 1233
a_attention_mask = torch.ones(a_batch_size, 320).long()
a_token_type_ids = torch.ones(a_batch_size, 320).long()

inputs = {
    'q_ids': q_ids,
    'a_input_ids': a_input_ids,
    'a_attention_mask': a_attention_mask,
    'a_token_type_ids': a_token_type_ids
}

inputs = {k: v.to('cuda') if config.use_gpu else v for k, v in inputs.items()}
output_model_path = './bert-base-uncase-ti-log-320head-s_n.onnx'
half_output_model_path = './bert-base-uncase-ti-log-320head-s_n.16.onnx'

start = time.time()
outputs = model(**inputs)
end = time.time()
print("PyTorch Inference time = ", end - start)

with torch.no_grad():
    q_names = {0: 'q_batch_size', 1: 'max_q_len'}
    a_names = {0: 'a_batch_size'}
    qa_names = {0: 'q_batch_size', 1: 'a_batch_size'}

    torch.onnx.export(model,  # model being run
                      (inputs['q_ids'],
                       inputs['a_input_ids'],  # model input (or a tuple for multiple inputs)
                       inputs['a_attention_mask'],
                       inputs['a_token_type_ids']),
                      output_model_path,  # where to save the model (can be a file or file-like object)
                      opset_version=11,  # the ONNX version to export the model to
                      do_constant_folding=True,  # whether to execute constant folding for optimization
                      input_names=['q_ids',
                                   'a_input_ids',  # the model's input names
                                   'a_attention_mask',
                                   'a_token_type_ids'],
                      output_names=['scores'],  # the model's output names
                      dynamic_axes={'q_ids': q_names,
                                    'a_input_ids': a_names,  # variable length axes
                                    'a_attention_mask': a_names,
                                    'a_token_type_ids': a_names,
                                    'scores': qa_names})
    print("Model exported at ", output_model_path)

onnx_model = onnxmltools.utils.load_model(output_model_path)
onnx_model = convert_float_to_float16(onnx_model)
onnxmltools.utils.save_model(onnx_model, half_output_model_path)
